local addonName, a = ...
local L = a.Localize
local s = SpellFlashAddon
local x = s.UpdatedVariables
local c = BittensGlobalTables.GetTable("BittensSpellFlashLibrary")

local GetCombatRatingBonus = GetCombatRatingBonus
local GetMasteryEffect = GetMasteryEffect
local GetRuneType = GetRuneType
local GetSpecialization = GetSpecialization
local GetSpellInfo = GetSpellInfo
local GetTime = GetTime
local GetTotemInfo = GetTotemInfo
local GetVersatility = GetVersatility
local GetVersatilityBonus = GetVersatilityBonus
local UnitAttackPower = UnitAttackPower
local UnitBuff = UnitBuff
local UnitExists = UnitExists
local UnitHealthMax = UnitHealthMax
local UnitLevel = UnitLevel
local UnitStat = UnitStat

local CR_VERSATILITY_DAMAGE_DONE = CR_VERSATILITY_DAMAGE_DONE

local select = select
local type = type
local print = print
local max = math.max


a.Costs = {}
function a.SetCost(blood, frost, unholy, death, name, bonusRP, rime, freeWithPerk)
   a.Costs[s.SpellName(c.GetID(name))] = {
      Blood = blood,
      Frost = frost,
      Unholy = unholy,
      Death = death,
      BonusRP = bonusRP,
      Rime = rime,
      FreeWithPerk = freeWithPerk and c.GetID(freeWithPerk) or nil
   }
end

a.SetCost(1, 0, 0, 0, "Blood Boil")
a.SetCost(0, 0, 0, 0, "Bone Shield")
a.SetCost(0, 1, 0, 0, "Chains of Ice")
a.SetCost(0, 0, 1, 0, "Dark Transformation", nil, false, "Enhanced Dark Transformation")
a.SetCost(0, 0, 1, 0, "Death and Decay")
a.SetCost(0, 0, 0, 1, "Death Siphon")
a.SetCost(0, 1, 1, 0, "Death Strike")
a.SetCost(0, 0, 0, 0, "Empower Rune Weapon", 25)
a.SetCost(1, 1, 0, 0, "Festering Strike")
a.SetCost(0, 0, 0, 0, "Horn of Winter")
a.SetCost(0, 1, 0, 0, "Howling Blast", nil, true)
a.SetCost(0, 1, 0, 0, "Icy Touch", nil, true)
a.SetCost(0, 1, 1, 0, "Obliterate")
a.SetCost(0, 1, 0, 0, "Pillar of Frost", nil, false, "Empowered Pillar of Frost")
a.SetCost(0, 0, 1, 0, "Plague Strike")
a.SetCost(1, 0, 0, 0, "Rune Tap")
a.SetCost(0, 0, 1, 0, "Scourge Strike")
a.SetCost(1, 0, 0, 0, "Strangulate")


local function runeAvailable(runeId, delay)
   if delay == nil then
      delay = 0
   end
   if a.Runes[runeId] <= delay then
      return 1
   end
end

local function fakeConsume(runeID, delay, needed, deathAvailable)
   if runeAvailable(runeID, delay) then
      if a.IsDeathRune(runeID) then
         deathAvailable = deathAvailable + 1
      elseif needed > 0 then
         needed = needed - 1
      end
   end
   return needed, deathAvailable
end

local function sufficientRunes(blood, frost, unholy, death, forbidDeath, noGCD)
   local deathAvailable = a.PendingDeathRunes
   local delay = 0
   if noGCD then
      delay = -c.GetBusyTime()
   end
   blood, deathAvailable = fakeConsume(1, delay, blood, deathAvailable)
   blood, deathAvailable = fakeConsume(2, delay, blood, deathAvailable)
   frost, deathAvailable = fakeConsume(5, delay, frost, deathAvailable)
   frost, deathAvailable = fakeConsume(6, delay, frost, deathAvailable)
   unholy, deathAvailable = fakeConsume(3, delay, unholy, deathAvailable)
   unholy, deathAvailable = fakeConsume(4, delay, unholy, deathAvailable)

   -- handle macroed BT, which can maybe generate a death rune when
   -- we hit the button for this action, prior to it landing.
   if c.GetOption("BTMacro") and c.HasTalent("Blood Tap") then
      if a.BloodCharges >= 5 and a.hasFullyDepleted(1, 2, 3, 4, 5, 6) then
         deathAvailable = deathAvailable + 2
      end
   end

   if forbidDeath then
      deathAvailable = 0
   end

   -- c.Debug("sufficientRunes", blood, frost, unholy, death, '=', blood+frost+unholy+death, '<=', deathAvailable, '|', blood+frost+unholy+death <= deathAvailable)
   return blood + frost + unholy + death <= deathAvailable
end

local function sufficientResources(id, noGCD)
   local name = s.SpellName(id)
   local cost = a.Costs[name]
   if cost then
      if cost.freeWithPerk and c.HasSpell(cost.freeWithPerk) then
         return true
      end

      local frost = cost.Rime and a.FreezingFog > 0 and 0 or cost.Frost
      return sufficientRunes(cost.Blood, frost, cost.Unholy, cost.Death, false, noGCD)
   else
-- c.Debug("Resources", name, s.SpellCost(name, SPELL_POWER_RUNIC_POWER), "<=", a.RP, "|", s.SpellCost(name) <= a.RP)
      return s.SpellCost(name, SPELL_POWER_RUNIC_POWER) <= a.RP
   end
end

local function masterOverride(z)
   if z.Melee then
      if not s.MeleeDistance() then
         return false
      end
   elseif not z.NoRangeCheck and s.SpellHasRange(z.ID) then
      if not s.SpellInRange(z.ID) then
         return false
      end
   end
   return (z.CheckFirst == nil or z:CheckFirst())
      and sufficientResources(z.ID)
end

-- Use this for all spells with rune or RP costs.  Spells added this way may
-- only use the CheckFirst and NoRangeCheck attributes.  All others will be
-- ignored.  Cooldowns will NOT be checked.
local function addSpell(name, tag, attributes)
   c.AddSpell(name, tag, attributes).Override = masterOverride
end

local function addOptionalSpell(name, tag, attributes)
   c.AddOptionalSpell(name, tag, attributes).Override = masterOverride
end

------------------------------------------------------------------------ common
function a.IsDeathRune(runeId)
   return GetRuneType(runeId) == 4
end

function a.BothRunesAvailable(runeName, delay)
   if runeName == "Blood" then
      return runeAvailable(1, delay) and runeAvailable(2, delay)
   elseif runeName == "Frost" then
      return runeAvailable(5, delay) and runeAvailable(6, delay)
   elseif runeName == "Unholy" then
      return runeAvailable(3, delay) and runeAvailable(4, delay)
   end
   print("Bad rune name: ", runeName)
end

local function isFullyDepleted(runeID)
   local otherID
   if runeID % 2 == 0 then
      otherID = runeID - 1
   else
      otherID = runeID + 1
   end
   local otherCD = a.Runes[otherID]
   return otherCD > 0 and a.Runes[runeID] > otherCD
end

function a.hasFullyDepleted(...)
   for i = 1, select("#", ...) do
      if isFullyDepleted(select(i, ...)) then
         return true
      end
   end
end

local function hasFrostFever(earlyRefresh)
   if c.IsCasting("Plague Leech") then
      return false
   end

   if c.HasTalent("Necrotic Plague") then
      return c.GetMyDebuffDuration("Necrotic Plague") > earlyRefresh
      or c.IsAuraPendingFor("Outbreak")
      or c.IsAuraPendingFor("Plague Strike")
      or c.IsAuraPendingFor("Icy Touch")
      or c.IsAuraPendingFor("Howling Blast")
      or c.HasBuff("Unholy Blight")
      or c.IsCasting("Unholy Blight")
      or (c.HasSpell("Ebon Plaguebringer")
         and c.IsAuraPendingFor("Plague Strike"))
   end

   return c.GetMyDebuffDuration("Frost Fever") > earlyRefresh
      or c.IsAuraPendingFor("Outbreak")
      or c.IsAuraPendingFor("Icy Touch")
      or c.IsAuraPendingFor("Howling Blast")
      or c.HasBuff("Unholy Blight")
      or c.IsCasting("Unholy Blight")
      or (c.HasSpell("Ebon Plaguebringer")
         and c.IsAuraPendingFor("Plague Strike"))
   -- @todo danielp 2014-10-18: check blood DK rules
end

local function hasBloodPlague(earlyRefresh)
   if c.IsCasting("Plague Leech") then
      return false
   end

   if c.HasTalent("Necrotic Plague") then
      return c.GetMyDebuffDuration("Necrotic Plague") > earlyRefresh
      or c.IsAuraPendingFor("Outbreak")
      or c.IsAuraPendingFor("Plague Strike")
      or c.IsAuraPendingFor("Icy Touch")
      or c.IsAuraPendingFor("Howling Blast")
      or c.HasBuff("Unholy Blight")
      or c.IsCasting("Unholy Blight")
      or (c.HasSpell("Ebon Plaguebringer")
         and c.IsAuraPendingFor("Plague Strike"))
   end

   return c.GetMyDebuffDuration("Blood Plague") > earlyRefresh
      or c.IsAuraPendingFor("Outbreak")
      or c.IsAuraPendingFor("Plague Strike")
      or GetTime() - a.BloodPlagueRefreshPending < .8
      or c.HasBuff("Unholy Blight")
      or c.IsCasting("Unholy Blight")
   -- @todo danielp 2014-10-18: check blood DK rules
end

local function hasBothDiseases(earlyRefresh)
   return hasFrostFever(earlyRefresh) and hasBloodPlague(earlyRefresh)
end

-- local function shouldStrikeForResources()
--    if c.HasTalent("Blood Tap") then
--       return a.BloodCharges <= 10
--          or c.GetBuffDuration("Blood Charge") < 5
--    elseif c.HasTalent("Runic Empowerment") then
--       return a.hasFullyDepleted(1, 2, 3, 4, 5, 6)
--    elseif c.HasTalent("Runic Corruption") then
--       return c.GetBuffDuration("Runic Corruption") == 0
--    end
-- end

local function soulReaperIsReady(delay)
   if type(delay) ~= "number" then
      delay = 0
   end

   return a.InExecute and
      GetTime() + c.GetBusyTime() + delay - a.LastSoulReaper >= 6 and
      s.MeleeDistance() and
      not c.IsCasting(
         "Soul Reaper - Frost",
         "Soul Reaper - Unholy",
         "Soul Reaper - Blood")
end

c.AddOptionalSpell("Empower Rune Weapon", nil, {
   NoGCD = true,
   FlashSize = s.FlashSizePercent() / 2,
})

c.AddSpell("Empower Rune Weapon", "for Death Strike", {
   NoGCD = true,
   CheckFirst = function()
      -- don't burn this just for saving the shield
      return a.ShouldDeathStrikeForHealth()
   end
})

c.AddOptionalSpell("Empower Rune Weapon", "when low", {
   NoGCD = true,
   FlashSize = s.FlashSizePercent() / 2,
   CheckFirst = function()
      return not (runeAvailable(1)
            or runeAvailable(2)
            or runeAvailable(5)
            or runeAvailable(6)
            or runeAvailable(3)
            or runeAvailable(4))
         and a.PendingDeathRunes == 0
         and a.RP < 32
   end
})

c.AddOptionalSpell("Blood Tap", nil, {
   NoGCD = true,
   CheckFirst = function()
      return a.BloodCharges >= 5
         and a.hasFullyDepleted(1, 2, 3, 4, 5, 6)
   end
})

c.AddOptionalSpell("Blood Tap", "at 11", {
   NoGCD = true,
   CheckFirst = function()
      return a.BloodCharges >= 11
         and a.hasFullyDepleted(1, 2, 3, 4, 5, 6)
   end
})

c.AddOptionalSpell("Blood Tap", "for Defile", {
   NoGCD = true,
   CheckFirst = function()
      return a.BloodCharges >= 5
         and a.hasFullyDepleted(1, 2, 3, 4, 5, 6) and
         c.HasTalent("Defile") and
         c.GetCooldown("Defile", false, 30) == 0
   end
})

addOptionalSpell("Death Siphon", nil, {
   CheckFirst = function()
      return c.GetHealthPercent("player") < 80
   end
})

c.AddSpell("Outbreak", nil, {
   Applies = { "Blood Plague", "Frost Fever" },
   CheckFirst = function()
      return not hasBothDiseases(0)
   end
})

c.AddSpell("Outbreak", "at 3", {
   CheckFirst = function()
      return c.GetCooldown("Outbreak") == 0 and not hasBothDiseases(3)
   end
})

c.AddSpell("Unholy Blight", nil, {
   Melee = true,
   CheckFirst = function()
      return c.GetCooldown("Unholy Blight") == 0 and not hasBothDiseases(0)
   end
})

c.AddSpell("Unholy Blight", "at 2", {
   Melee = true,
   CheckFirst = function()
      return c.GetCooldown("Unholy Blight") == 0 and not hasBothDiseases(2)
   end
})

c.AddSpell("Unholy Blight", "at 3", {
   Melee = true,
   CheckFirst = function()
      return c.GetCooldown("Unholy Blight") == 0 and not hasBothDiseases(3)
   end
})

local function canPlagueLeech()
   if not hasBothDiseases(0) then
      return false
   end

   local spec = GetSpecialization()
   if spec == 3 then -- unholy spec
      return a.hasFullyDepleted(1, 2, 5, 6) -- blood & frost runes
   else -- blood or frost spec
      return a.hasFullyDepleted(5, 6, 3, 4) -- frost & unholy runes
   end
end

c.AddSpell("Plague Leech", nil, {
   CheckFirst = canPlagueLeech,
})

c.AddSpell("Plague Leech", "for Death Strike", {
   CheckFirst = function()
      return canPlagueLeech() and (
         a.ShouldDeathStrikeForHealth() or
            a.shouldDeathStrikeForShield())
   end
})

c.AddSpell("Plague Leech", "at 1", {
   CheckFirst = function()
      return canPlagueLeech() and not hasBothDiseases(1)
   end
})

c.AddSpell("Plague Leech", "at 2", {
   CheckFirst = function()
      return canPlagueLeech() and not hasBothDiseases(2)
   end
})

c.AddSpell("Plague Leech", "at 3", {
   CheckFirst = function()
      return canPlagueLeech() and not hasBothDiseases(3)
   end
})

c.AddSpell("Plague Leech", "if Outbreak", {
   CheckFirst = function()
      return canPlagueLeech() and c.GetCooldown("Outbreak", false, 60) < 1
   end
})

addSpell("Plague Strike", "for Blood Plague", {
   Applies = { "Blood Plague" },
   CheckFirst = function()
      return not hasBloodPlague(0)
   end
})

c.AddOptionalSpell("Raise Dead", nil, {
   NoRangeCheck = 1,
   CheckFirst = function()
      return not UnitExists("pet")
   end
})

addSpell("Death and Decay", nil, {
   NoRangeCheck = true,
   CheckFirst = function()
      return c.GetCooldown("Death and Decay", false, 30) == 0
   end
})

c.AddOptionalSpell("Horn of Winter", "for Buff, Optional", {
   Cooldown = 20,
   CheckFirst = function()
      return c.RaidBuffNeeded(c.ATTACK_POWER_BUFFS)
   end
})

c.AddInterrupt("Mind Freeze", nil, {
   NoGCD = true,
})

c.AddInterrupt("Asphyxiate")

c.AddInterrupt("Strangulate", nil, {
   NoGCD = true,
})

addSpell("Defile", nil, {
   Melee = true,
})

------------------------------------------------------------------------- Blood
local function canBloodTapForDeathStrike()
   local frost = 0
   local unholy = 0
   local death = a.PendingDeathRunes
   if runeAvailable(5) then
      if a.IsDeathRune(5) then
         death = death + 1
      else
         frost = frost + 1
      end
   end
   if runeAvailable(6) then
      if a.IsDeathRune(6) then
         death = death + 1
      else
         frost = frost + 1
      end
   end
   if runeAvailable(3) then
      if a.IsDeathRune(3) then
         death = death + 1
      else
         unholy = unholy + 1
      end
   end
   if runeAvailable(4) then
      if a.IsDeathRune(4) then
         death = death + 1
      else
         unholy = unholy + 1
      end
   end
   return (frost + unholy + death == 1)
      or (death == 0 and (
            (frost == 0 and unholy > 0)
         or (unholy == 0 and frost > 0)))
end

-- Estimated amount of Blood Shield
-- Based on http://forums.wowace.com/showpost.php?p=329939&postcount=2
-- Uses data from http://us.battle.net/wow/en/forum/topic/13087818929#2
-- ItemScaling value by player level:
local item_scaling_by_level = {
   -- ten levels per line
     3,   3,   4,   4,   5,   6,   6,   7,   7,   8,
     8,   9,   9,  10,  10,  11,  11,  12,  12,  13,
    13,  14,  14,  15,  15,  16,  16,  17,  17,  18,
    18,  19,  19,  20,  20,  21,  21,  22,  22,  23,
    23,  24,  24,  25,  25,  26,  26,  27,  27,  28,
    28,  29,  29,  30,  30,  31,  31,  32,  32,  32,
    35,  37,  39,  39,  40,  40,  41,  44,  44,  44,
    44,  44,  45,  46,  49,  49,  50,  50,  51,  51,
    52,  52,  54,  56,  57,  60,  61,  62,  64,  67,
   101, 118, 139, 162, 190, 225, 234, 242, 252, 261}

local function getDeathStrikeHeal(forceUseScentStacks)
   --Death strike tooltip heal without versatility multiplier
   local baseAP, plusAP, minusAP = UnitAttackPower("player")
   local base_heal = (baseAP + plusAP - minusAP) * 5

   --Clean versatility value from items
   local mastery = 0.01 * GetMasteryEffect("player")

   --versatility as % multiplier; calculation based on "damage done",
   --which is currently == to "healing done", and this is stolen from
   --PaperDollFrame.lua from the Blizzard FrameXML code.
   local versatility = 1 + (
      0.01 * (GetCombatRatingBonus(CR_VERSATILITY_DAMAGE_DONE) +
            GetVersatilityBonus(CR_VERSATILITY_DAMAGE_DONE)))


   --Scent of Blood multiplier
   local SoB = 1 + (c.GetBuffStack("Scent of Blood") * 0.2)
   if not forceUseScentStacks and c.IsCasting("Death Strike") then
      SoB = 1
   end

   --Base resolve out of combat
   local stamina = UnitStat("player",3)
   local resolve_base = stamina / 250.0 / item_scaling_by_level[UnitLevel("Player")]
   local resolve_buff = select(17, UnitBuff("player", GetSpellInfo(c.GetID("Resolve")))) or 0

   --Resolve: http://us.battle.net/wow/en/forum/topic/13087818929#2
   local resolve = 1 + (resolve_buff / stamina / 60 * 0.25 + resolve_base)

   -- Dark Succor doubles DS healing...
   local dark_succor = c.HasBuff("Dark Succor") and 2 or 1

   -- finally, our healing estimate.
   return (((((base_heal * versatility) * SoB) * mastery) * resolve) * dark_succor)
end

function a.ShouldDeathStrikeForHealth()
   local health = c.GetHealth("player") + getDeathStrikeHeal()
   if c.IsCasting("Death Strike") then
      health = health + getDeathStrikeHeal(true)
   end

   -- allow up to one percent max health overheal
   return health <= (UnitHealthMax("player") * 1.01)
end

function a.shouldDeathStrikeForShield()
   return c.GetBuffDuration("Blood Shield") < 2
      and not c.IsCasting("Death Strike")
end

c.AddOptionalSpell("Blood Presence", nil, {
   Type = "form"
})

addSpell("Death Strike")

addSpell("Death Strike", "for Health", {
   CheckFirst = a.ShouldDeathStrikeForHealth
})

addSpell("Death Strike", "to Save Shield", {
    CheckFirst = function()
       return a.shouldDeathStrikeForShield()
    end
})

addSpell("Death Strike", "if Two Available", {
   CheckFirst = function()
      local avail = a.PendingDeathRunes
      if a.IsDeathRune(1) and runeAvailable(1, 1) then
         avail = avail + 1
      end
      if a.IsDeathRune(2) and runeAvailable(2, 1) then
         avail = avail + 1
      end
      if runeAvailable(5, 1) then
         avail = avail + 1
      end
      if runeAvailable(6, 1) then
         avail = avail + 1
      end
      if runeAvailable(3, 1) then
         avail = avail + 1
      end
      if runeAvailable(4, 1) then
         avail = avail + 1
      end
      if a.BloodCharges >= 10 then
         avail = avail + 2
      elseif a.BloodCharges >= 5 then
         avail = avail + 1
      end
      return avail >= 4
   end
})

c.AddOptionalSpell("Blood Tap", "for Death Strike", {
   NoGCD = true,
   Melee = true,
   CheckFirst = function()
      return a.BloodCharges >= 5
         and (a.ShouldDeathStrikeForHealth() or a.shouldDeathStrikeForShield())
         and canBloodTapForDeathStrike()
   end
})

addSpell("Death Siphon", "for Health", {
   CheckFirst = function()
      -- use the same criteria as for Death Strike, so we know Death Siphon
      -- is only used when Death Strike cannot be
      return a.ShouldDeathStrikeForHealth()
   end
})

c.AddSpell("Outbreak", "Early", {
   CheckFirst = function()
      return not hasBothDiseases(5)
   end
})

c.AddSpell("Outbreak", "Early, Glyphed", {
   CheckFirst = function()
      return c.HasGlyph("Outbreak") and not hasBothDiseases(5)
   end
})

addSpell("Death Coil")

addSpell("Death Coil", "1 Rune to cap", {
    CheckFirst = function() return a.RP > 90 end
})

addSpell("Death Coil", "2 Runes to cap", {
    CheckFirst = function() return a.RP > 80 end
})

addSpell("Death and Decay", "Free", {
   NoRangeCheck = true,
   CheckFirst = function()
      return a.CrimsonScourge
         and c.GetCooldown("Death and Decay", false, 30) == 0
   end
})

c.AddSpell("Blood Boil", "BB", {
   Melee = true,
   CheckFirst = function()
      -- blood runes only, forbid death runes
      return sufficientRunes(2, 0, 0, 0, true)
   end
})

c.AddSpell("Blood Boil", "B, if safe", {
   Melee = true,
   CheckFirst = function()
      -- Blood runes only, forbid death runes
      if not sufficientRunes(1, 0, 0, 0, true) then
         return false
      end

      -- only if we are not using RE, for which we should keep our
      -- second B rune active at all times.
      if c.HasTalent("Runic Empowerment") then
         return false
      end

      -- If we are in execute range, we want to hold our B for SR,
      -- too, so ensure that we will have a B rune ready by the time
      -- SR rolls around.
      --
      -- sanity check: is this gonna get messed up if things regen
      -- as death runes, which we don't want to spend on SR?
      -- I think we are actually safe on that front.
      if not a.InExecute then
         return true
      end

      if not soulReaperIsReady(max(a.Runes[1], a.Runes[2])) then
         return false
      end

      -- wow, everything passed!
      return true
   end
})

c.AddSpell("Blood Boil", "for diseases", {
   Melee = true,
   CheckFirst = function()
      return hasBothDiseases(0) and not hasBothDiseases(3)
   end
})

c.AddSpell("Blood Boil", "Free", {
   Melee = true,
   Override = function()
      return a.CrimsonScourge
   end
})

c.AddSpell("Soul Reaper - Blood", "B", {
   CheckFirst = function()
      return soulReaperIsReady() and sufficientRunes(1, 0, 0, 0, true)
   end
})

c.AddOptionalSpell("Rune Tap", nil, {
   NoGCD = true,
   Score = 3,
})

c.AddOptionalSpell("Rune Tap", "with more than one charge", {
   NoGCD = true,
   Score = 3,
   CheckFirst = function()
      return c.GetChargeInfo("Rune Tap") > 1
   end
})

c.AddOptionalSpell("Dancing Rune Weapon", nil, {
   NoGCD = true,
   Melee = true,
   CheckFirst = function()
      return not c.WearingSet(4, "BloodT16")
   end,
})

c.AddOptionalSpell("Dancing Rune Weapon", "Prime", {
   NoGCD = true,
   Melee = true,
   CheckFirst = function()
      return (c.WearingSet(4, "BloodT16")
            and (not (runeAvailable(1)
               or runeAvailable(2)
               or runeAvailable(3)
               or runeAvailable(4))))
         or c.InDamageMode()
   end,
})

c.AddOptionalSpell("Bone Shield", nil, {
   CheckFirst = function()
      if s.InCombat() then
         return not c.HasBuff("Bone Shield")
      elseif x.EnemyDetected then
         return c.GetBuffDuration("Bone Shield") < 60
            or s.BuffStack(c.GetID("Bone Shield"), "player") < 6
      end
   end
})

c.AddOptionalSpell("Vampiric Blood", nil, {
   NoGCD = true,
   CheckFirst = function()
      return not c.IsSolo() or sufficientResources(c.GetID("Death Strike"))
   end,
   ShouldHold = function()
      -- if we would self-heal, this is probably appropriate.
      return not a.ShouldDeathStrikeForHealth()
   end,
})

c.AddOptionalSpell("Death Pact", nil, {
   NoRangeCheck = 1,
   ShouldHold = function()
      return c.GetHealthPercent("player") > 45
   end
})

c.AddOptionalSpell("Conversion", nil, {
   NoGCD = true,
   Buff = "Conversion",
   CheckFirst = function()
      return a.RP > 50 and not c.HasBuff("Conversion")
   end,
   ShouldHold = function()
      return c.GetHealthPercent("player") > 80
   end
})

c.AddOptionalSpell("Conversion", "Cancel", {
   NoGCD = true,
   FlashColor = "red",
   CheckFirst = function()
      return c.HasBuff("Conversion", true)
         and not a.ShouldDeathStrikeForHealth()
   end
})

c.AddOptionalSpell("Icebound Fortitude", "Unglyphed", {
   Score = 2,
   NoGCD = true,
   Enabled = function()
      return not c.HasGlyph("Icebound Fortitude")
   end,
})

c.AddOptionalSpell("Icebound Fortitude", "Glyphed", {
   Score = 2,
   NoGCD = true,
   Enabled = function()
      return c.HasGlyph("Icebound Fortitude")
   end,
})

c.AddTaunt("Dark Command", nil, {
   NoGCD = true
})

c.AddTaunt("Death Grip", nil, {
   NoGCD = true
})

------------------------------------------------------------------------- Frost
c.AddOptionalSpell("Frost Presence", nil, { Type = "form" })

c.AddOptionalSpell("Pillar of Frost", nil, {
   NoGCD = true,
})

addSpell("Frost Strike", nil, {
   Melee = true,
})

addSpell("Frost Strike", "under KM", {
   Melee = true,
   CheckFirst = function()
      return a.KillingMachine
   end
})

addSpell("Frost Strike", "for RE", {
   Melee = true,
   CheckFirst = function()
      return c.HasTalent("Runic Empowerment") and
         a.hasFullyDepleted(1, 2, 3, 4, 5, 6) and
         (not a.Killing_machine or not sufficientResources(c.GetID("Obliterate"), false))
   end
})

addSpell("Frost Strike", "for BT", {
   Melee = true,
   CheckFirst = function()
      return c.HasTalent("Blood Tap") and
         a.BloodCharges <= 10 and
         (not a.Killing_machine or not sufficientResources(c.GetID("Obliterate"), false))
   end
})

addSpell("Frost Strike", "for OB cap", {
   Melee = true,
   CheckFirst = function()
      return a.RP > 76 and
         sufficientResources(c.GetID("Obliterate"), false)
   end
})

addSpell("Frost Strike", "for HB cap", {
   Melee = true,
   CheckFirst = function()
      return a.RP > 88 and
         sufficientResources(c.GetID("Howling Blast"), false)
   end
})

addSpell("Frost Strike", "for PS cap", {
   Melee = true,
   CheckFirst = function()
      return a.RP > 88 and
         sufficientResources(c.GetID("Plague Strike"), false)
   end
})

addSpell("Soul Reaper - Frost", nil, {
   CheckFirst = soulReaperIsReady,
})

c.AddSpell("Blood Tap", "for Soul Reaper - Frost", {
   NoGCD = true,
   Melee = true,
   FlashID = { "Blood Tap", "Soul Reaper - Frost" },
   CheckFirst = function()
      return soulReaperIsReady()
         and a.BloodCharges >= 5
         and a.hasFullyDepleted(1, 2, 3, 4, 5, 6)
   end
})

c.AddSpell("Blood Tap", "for OB KM", {
   NoGCD = true,
   Melee = true,
   FlashID = { "Blood Tap", "Obliterate" },
   CheckFirst = function()
      return a.KillingMachine
         and a.BloodCharges >= 5
         and (a.PendingDeathRunes > 0
            or runeAvailable(1)
            or runeAvailable(2)
            or runeAvailable(5)
            or runeAvailable(6)
            or runeAvailable(3)
            or runeAvailable(4))
   end
})

c.AddOptionalSpell("Blood Tap", "at 8 or Non-Execute", {
   NoGCD = true,
   CheckFirst = function()
      if not a.hasFullyDepleted(1, 2, 3, 4, 5, 6) then
         return false
      end

      if a.BloodCharges >= 8 then
         return true
      end

      return a.BloodCharges >= 5 and not a.InExecute
   end
})

addSpell("Howling Blast")

addSpell("Howling Blast", "for Frost Fever", {
   Applies = { "Frost Fever" },
   CheckFirst = function()
      return not hasFrostFever(0)
   end
})

addSpell("Howling Blast", "under Freezing Fog", {
   CheckFirst = function()
      return a.FreezingFog > 0
   end
})

addSpell("Obliterate", nil, {
})

local function checkObliterateUnholyWithoutKillingMachine()
   return ((runeAvailable(3) and not a.IsDeathRune(3))
         or (runeAvailable(4) and not a.IsDeathRune(4)))
      and not a.KillingMachine
end

addSpell("Obliterate", "U w/out KM", {
   CheckFirst = checkObliterateUnholyWithoutKillingMachine
})

addSpell("Frost Strike", "for Oblit cap no KM", {
   Melee = true,
   CheckFirst = function()
      return a.RP > 76 and checkObliterateUnholyWithoutKillingMachine()
   end
})


addSpell("Obliterate", "under KM", {
   CheckFirst = function()
      return a.KillingMachine
   end
})

addSpell("Obliterate", "BB or FF or UU", {
   CheckFirst = function()
      return (a.BothRunesAvailable("Blood", 1)
            or a.BothRunesAvailable("Frost", 1)
            or a.BothRunesAvailable("Unholy", 1))
   end
})

addSpell("Death and Decay", "U", {
   NoRangeCheck = true,
   CheckFirst = function()
      return c.GetCooldown("Death and Decay", false, 30) == 0
         and sufficientRunes(0, 0, 1, 0, true)
   end
})

addOptionalSpell("Death and Decay", "U, optional", {
   NoRangeCheck = true,
   CheckFirst = function()
      return c.GetCooldown("Death and Decay", false, 30) == 0
         and sufficientRunes(0, 0, 1, 0, true)
   end
})

addSpell("Plague Strike", "U", {
   CheckFirst = function()
      return sufficientRunes(0, 0, 1, 0, true)
   end
})

addSpell("Plague Strike", "UU", {
   CheckFirst = function()
      return sufficientRunes(0, 0, 2, 0, true)
   end
})

------------------------------------------------------------------------ Unholy
c.AddOptionalSpell("Unholy Presence", nil, {
   Type = "form",
})

addSpell("Soul Reaper - Unholy", nil, {
   CheckFirst = soulReaperIsReady,
})

c.AddSpell("Blood Tap", "for Soul Reaper - Unholy", {
--      NoGCD = true,
   Melee = true,
   FlashID = { "Blood Tap", "Soul Reaper - Unholy" },
   CheckFirst = function()
      return soulReaperIsReady()
         and a.BloodCharges >= 5
         and a.hasFullyDepleted(1, 2, 3, 4, 5, 6)
   end
})

addSpell("Plague Strike", "for Both Diseases", {
   CheckFirst = function()
      return not hasBothDiseases(0)
   end
})

addOptionalSpell("Summon Gargoyle", nil, {
   CheckFirst = function()
      return c.GetCooldown("Summon Gargoyle") == 0
   end
})

local function shouldDT()
   if s.Buff(c.GetID("Dark Transformation"), "pet")
      or c.IsAuraPendingFor("Dark Transformation") then

      return false
   end

   local stacks = s.BuffStack(c.GetID("Shadow Infusion"), "pet")
   if a.ShadowInfusionPending or c.IsCasting("Death Coil") then
      stacks = stacks + 1
   end
   return stacks >= 5
end

addSpell("Dark Transformation", nil, {
   CheckFirst = shouldDT,
})

c.AddSpell("Blood Tap", "for Dark Transformation", {
--      NoGCD = true,
   FlashID = { "Blood Tap", "Dark Transformation" },
   CheckFirst = function()
      return a.BloodCharges >= 5
         and a.hasFullyDepleted(1, 2, 3, 4, 5, 6)
         and shouldDT()
   end
})

addSpell("Death Coil", "at 90", {
   CheckFirst = function()
      return a.RP > 90
   end
})

addSpell("Death and Decay", "UU", {
   NoRangeCheck = true,
   CheckFirst = function()
      return c.GetCooldown("Death and Decay", false, 30) == 0
         and a.BothRunesAvailable("Unholy", 1)
   end
})

addSpell("Death and Decay", "unless Soul Reaper", {
   NoRangeCheck = true,
   CheckFirst = function()
      return c.GetCooldown("Death and Decay", false, 30) == 0
         and (not a.InExecute
            or not soulReaperIsReady(1)
            or sufficientRunes(0, 0, 2, 0, false))
   end
})

c.AddOptionalSpell("Blood Tap", "for D&D", {
--      NoGCD = true,
   FlashID = { "Blood Tap", "Death and Decay" },
   CheckFirst = function()
      return a.BloodCharges >= 5
         and c.GetCooldown("Death and Decay") == 0
         and not sufficientResources(c.GetID("Death and Decay"))
         and not c.IsCasting("Death and Decay")
   end
})

addSpell("Scourge Strike", "UU", {
   CheckFirst = function()
      return a.BothRunesAvailable("Unholy", 1)
   end
})

addSpell("Scourge Strike", "unless Soul Reaper", {
   CheckFirst = function()
      return not a.InExecute
         or not soulReaperIsReady(1)
         or sufficientRunes(0, 0, 2, 0, false)
   end
})

addSpell("Festering Strike", nil, {
   CheckFirst = function()
      return sufficientRunes(1, 1, 0, 0, true)
   end
})

addSpell("Festering Strike", "BBFF", {
   CheckFirst = function()
      return a.BothRunesAvailable("Blood", 1)
         or a.BothRunesAvailable("Frost", 1)
   end
})

addSpell("Death Coil", "under Sudden Doom or for Dark Transformation", {
   CheckFirst = function()
      if c.HasBuff("Sudden Doom") and not c.IsCasting("Death Coil") then
         return true
      end

      if GetTime() - a.DTCast < .8
         or s.BuffDuration(c.GetID("Dark Transformation"), "pet")
            > c.GetBusyTime()
         or c.IsCasting("Dark Transformation") then

         return false
      end

      -- make sure unholy runes stay on cooldown?
      local runes = a.PendingDeathRunes
      if runeAvailable(1, 1) and a.IsDeathRune(1) then
         runes = runes + 1
      end
      if runeAvailable(2, 1) and a.IsDeathRune(2) then
         runes = runes + 1
      end
      if runeAvailable(5, 1) and a.IsDeathRune(5) then
         runes = runes + 1
      end
      if runeAvailable(6, 1) and a.IsDeathRune(6) then
         runes = runes + 1
      end
      if runeAvailable(3, 1) then
         runes = runes + 1
      end
      if runeAvailable(4, 1) then
         runes = runes + 1
      end
      return runes <= 1
   end
})

addSpell("Death Coil", "unless Gargoyle or Dark Transformation", {
   CheckFirst = function()
      local dt
      if c.IsAuraPendingFor("Dark Transformation") then
         dt = 30
      else
         dt = s.BuffDuration(c.GetID("Dark Transformation"), "pet")
            - c.GetBusyTime()
         if dt <= 0 then
            return true
         end
      end

      return dt > 8
         and (c.GetCooldown("Summon Gargoyle") > 8
            or c.IsCasting("Summon Gargoyle"))
   end
})

c.AddOptionalSpell("Blood Tap", "at 8 or for Dark Transformation", {
   NoGCD = true,
   CheckFirst = function()
      if a.InExecute or not a.hasFullyDepleted(1, 2, 3, 4, 5, 6) then
         return false
      end

      if a.BloodCharges >= 8 then
         return true
      end

      return a.BloodCharges >= 5
         and s.BuffDuration(c.GetID("Dark Transformation"), "pet")
            <= c.GetBusyTime()
   end
})
